
require 'set'

module Ridge
	@@encoding = nil
	def self.encoding()
		@@encoding
	end

	def self.encoding=(val)
		@@encoding = val
	end
	
	module GUI
		Widgets = {}

		class Widget
			def initialize(rc, properties, children=[])
				@resource = rc
				@properties = properties
				@children = children
				@id = properties['id']
				@events = EventDispatch.new()
				@widgets = {}
				@children.each do |child|
					@widgets[child.id] = child if child.id
					@widgets.update(child.widgets)
				end
			end
			attr :resource
			attr :properties
			attr :id
			attr :events
			attr :widgets
			attr :children
			
			def widget(id)
				return self.widgets[id]
			end
			
			def native()
				nil
			end

			def attach_to(parent)
			end
		end

		class Bind
			BindType = {'integer'=>:to_i, 'string'=>:to_s, 'float'=>:to_f}

			def bind_widgets(widget)
				if widget.properties['bind'] then
					widget.events.add_handler('on_change', self.method(:on_change))
					@binds.push([widget.properties['bind'], widget])
				end
				if widget.respond_to?(:children) then
					widget.children.each do |child|
						bind_widgets(child)
					end
				end
			end
			private :bind_widgets
			
			def initialize(widget, target)
				@updating = false
				@target = target

				@binds = []
				bind_widgets(widget)
			end
			attr :target, true

			def on_change(sender, args)
				return if @updating
				return if @target.nil?
				@updating = true
				value = args['value']
				bindtype = sender.properties['bindtype']
				if bindtype && BindType.include?(bindtype) then
					value = value.__send__(BindType[bindtype])
				end

				bind = sender.properties['bind']
				begin
					@target.__send__(bind+'=', value)
				rescue NoMethodError
					@target.__send__(bind, value)
				end
				@updating = false
			end

			def update()
				return if @updating
				return if @target.nil?
				@updating = true
				@binds.each do |bind, widget|
					value = @target.__send__(bind)
					if widget.respond_to?(:value_type) then
						value = value.__send__(BindType[widget.value_type])
					end
					widget.value = value
				end
				@updating = false
			end
		end
	end

	module Graphics
		Images = {}

		class Image
			def initialize(rc, properties, data=nil)
				@resource = rc
				@data = data
				@id = properties['id']
			end
			attr :resource
			attr :element
			attr :id
			
			def native()
				nil
			end
		end
	end
	
  module Boolean
    def self.parse(val)
      unless val.nil? or val.kind_of?(TrueClass) or val.kind_of?(FalseClass) then
        case val.to_s
        when /1|yes|true/i
          return true
        when /0|no|false/i
          return false
        else
          return nil
        end
      else
        return val
      end
    end
  end

	class Length
		@@resolution = 72.0 # pixels/inch
		
		def initialize(val)
			if val.kind_of?(String) then
				md = /([+-]?\d*(\.\d+)?)(in|pt|cm|mm|px)?/.match(val)
				raise "Unknown format:#{str}" if md.nil?
        @len = md[1].to_f
				case md[3]
				when 'in'
					@len *= 25.4
				when 'pt'
					@len = @len * 25.4 / 72.0
				when 'cm'
					@len *= 10
				when 'mm'
					@len = @len
				when 'px'
					@len = @len * 25.4 / @@resolution
				else #default to px
					@len = @len * 25.4 / @@resolution
				end
      elsif val.kind_of?(Numeric) then
        #defualt to px
        @len = val.to_f * 25.4 / @@resolution
      else
				@len = 0.0
			end
		end

		def to_in()
			@len / 25.4
		end

		def to_pt()
			@len = @len * 25.4 / 72.0
		end

		def to_cm()
			@len / 10.0
		end
		
		def to_mm()
			@len
		end

		def to_px()
			(@len * @@resolution / 25.4).to_i
		end

		def self.resolution()
			@@resolution
		end

		def self.resolution=(newval)
			@@resolution = newval.to_f
		end

		def self.to_px(val)
			self.new(val).to_px
		end
	end
	
	class EventDispatch
		def initialize()
			@events = {}
			@dispatching = Set.new()
		end

		def add_handler(name, handler=nil)
			handler = Proc.new() if handler.nil?
			@events[name] = [] unless @events.include?(name)
			@events[name].push(handler)
		end

		def remove_handler(name, handler=nil)
			@events.delete(name) if handler.nil?
			if @events.include?(name) then
				@events[name].delete(handler)
			end
		end
		
		def fire(name, *args)
			if !@dispatching.include?(name) && @events.include?(name) then
				@dispatching.add(name)
				@events[name].each do |handler|
					handler.call(*args)
				end
				@dispatching.delete(name)
			end
		end
	end
			
	class ResourceReader
		def initialize(encoding=nil)
			@encoding = encoding
		end

		def parse(stream)
			raise NotImplementedError
		end

		def encoding()
			@encoding
		end
	end

	class ResourceElement
		def initialize(name, properties, children, texts)
			@name = name
			@children = children
			@properties = properties
			@texts = texts
		end

		def children()
			@children
		end

		def texts()
			@texts
		end

		def find(id)
			if @properties.include?('id') and @properties['id']==id then
				return self
			else
				@children.each do |child|
					res = child.find(id)
					return res if res 
				end
				return nil
			end
		end
		
		def properties()
			@properties
		end

		def name()
			@name
		end
	end

	class ResourceElementRoot < ResourceElement
		def initialize(name, properties, children, texts)
			super
			@element_cache = {}
		end

		def find(id)
			if @element_cache.include?(id) then
				return @element_cache[id]
			else
				res = super
				@element_cache[id] = res
				return res
			end
		end
	end
	
	class Resource
		Reader = {}
		
		def initialize(stream, reader=REXMLResourceReader.new)
			@root = reader.parse(stream)
			@events = EventDispatch.new()
			@instances = []
			Length.resolution = @root.properties['resolution'].to_i if @root.properties['resolution']
		end
		attr :events

		def self.open(file, reader=nil)
			if not reader then
				ext = File.extname(file).sub(/^\./, '')
				Reader.each do |key, klass|
					reader = klass.new if ext===key
				end
				raise "Unknown file format:#{ext}" if reader.nil?
			end
			res = nil
			File.open(file, 'r') do |f|
				res = self.new(f, reader)
			end
			return res
		end

		def widget(id)
			elem = @root.find(id)
			raise "'#{id}' not found" if elem.nil?
			return WidgetResource.new(self, elem)
		end

		def image(id)
			elem = @root.find(id)
			raise "'#{id}' not found" if elem.nil?
			return ImageResource.new(self, elem)
		end
		
		def string(id)
			elem = @root.find(id)
			raise "'#{id}' not found" if elem.nil?
			raise "Unknown string resource '#{elem.name}'" if elem.name!='string'
			return elem.texts.join('')
		end

		class WidgetResource
			def initialize(rc, elem)
				@element = elem
				@resource = rc
			end

			def new()
				raise "Unknown widget '#{@element.name}'" if GUI::Widgets[@element.name].nil?
				children = []
				if @element.children then
					@element.children.each do |child|
						children.push(WidgetResource.new(@resource, child).new())
					end
				end
				GUI::Widgets[@element.name].new(@resource, @element.properties, children)
			end
		end

		class ImageResource
			def initialize(rc, elem)
				@element = elem
				@resource = rc
			end

			def new()
				raise "Unknown image '#{@element.name}'" if Graphics::Images[@element.name].nil?
				Graphics::Images[@element.name].new(@resource, @element.properties, nil)
			end
		end
	end

end

#Default reader
require 'ridge/reader/rexml'

